from typing import List
import numpy as np
from scipy import signal
from scipy.fftpack import fft


def fourier(epochized_data: List[int or float], sampling_freq: int or float) -> List[np.ndarray]:
	"""

	Args:
		epochized_data:
		sampling_freq:

	Returns:

	"""
	return_list = []
	T = 1 / sampling_freq
	for i in epochized_data:
		yf = fft(i)
		yf = np.array(yf)
		return_list.append(yf)
	return return_list



from typing import List
import pandas as pd
from scipy.io import loadmat


def load_mat(file: str) -> (List[int or float], pd.DataFrame):
	# If mat files all have the same format then hardcode how to insert the code into the dataframe and signals list
	data = loadmat(file)

	return



import matplotlib.pyplot as plt
from typing import Dict


def _pltcolor(lst):
    """
    Bent Ari gave us this code and this is based on his Bsc project
    Args:
        lst:

    Returns:

    """
    cols = []
    for num in lst:
        if num == 0:
            cols.append('fuchsia')  # Wake
        elif num == -1:
            cols.append('blue')  # N1
        elif num == -2:
            cols.append('aqua')  # N
        elif num == -3:
            cols.append('lime')  # N3
        elif num == 5:
            cols.append('orange')  # REM
        elif num == 12:
            cols.append('dodgerblue')  # N1 or N2 (both)
        elif num == -10:
            cols.append('red')
        else:
            cols.append('red')  # other
        return cols


def plot_distrib_pies(distrib_dict: Dict, sleep_stage_map: Dict, pct=False):
    """
    Bent Ari gave us this code and this is based his Bsc project
    Args:
        distrib_dict:
        sleep_stage_map:
        pct:
    """
    f, axs = plt.subplots(10, 6, figsize=(35, 35))
    for i in range(10):
        curr_primarykeys = list(distrib_dict[f'Scorer {i + 1}'].keys())
        curr_primarykeys.sort()
        for j in range(len(curr_primarykeys)):
            key = curr_primarykeys[j]
            labels = []
            sizes = []
            for x, y in distrib_dict[f'Scorer {i + 1}'][key].items():
                labels.append(x)
                sizes.append(y)
            # Plot
            cols = _pltcolor(labels)
            if pct:
                axs[i, j].pie(sizes, labels=labels, autopct='%1.1f%%', colors=cols)
            else:
                axs[i, j].pie(sizes, labels=labels, autopct=autopct_format(sizes), colors=cols)
            axs[i, j].set_title(f"Followup of {sleep_stage_map[key]}")
            axs[i, j].legend([sleep_stage_map[num] for num in labels], bbox_to_anchor=(1.05, 1), loc='upper left',
                             borderaxespad=0., prop={'size': 10})
            axs[i, j].text(0.95, 0.1, f'Switches: {sum(sizes):,}', horizontalalignment='center',
                           verticalalignment='center', transform=axs[i, j].transAxes)



import time
import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np


def plot_hypnogram(
		stages,
		labeldict=None,
		title=None,
		epochlen=30,
		ax=None,
		verbose=True,
		xlabel=True,
		ylabel=True,
		**kwargs, ):
	"""
	Plot a hypnogram, the flexible way.
	A labeldict should give a mapping which integer belongs to which class
	E.g labeldict = {0: 'Wake', 4:'REM', 1:'S1', 2:'S2', 3:'SWS'}
	or {0:'Wake', 1:'Sleep', 2:'Sleep', 3:'Sleep', 4:'Sleep', 5:'Artefact'}
	The order of the labels on the plot will be determined by the order of the dictionary.
	E.g.  {0:'Wake', 1:'REM', 2:'NREM'}  will plot Wake on top, then REM, then NREM
	This dictionary can be infered automatically from the numbers that are present
	in the hypnogram but this functionality does not cover all cases.
	:param stages: An array with different stages annotated as integers
	:param labeldict: An enumeration of labels that correspond to the integers of stages
	:param title: Title of the window
	:param epochlen: How many seconds is one epoch in this annotation
	:param ax: the axis in which we plot
	:param verbose: Print stuff or not.
	:param xlabel: Display xlabel ('Time after record start')
	:param ylabel: Display ylabel ('Sleep Stage')
	:param kwargs: additional arguments passed to plt.plot(), e.g. c='red'
	"""

	if labeldict is None:
		if np.max(stages) == 1 and np.min(stages) == 0:
			labeldict = {0: 'W', 1: 'S'}
		elif np.max(stages) == 2 and np.min(stages) == 0:
			labeldict = {0: 'W', 2: 'REM', 1: 'NREM'}
		elif np.max(stages) == 4 and np.min(stages) == 0:
			if 1 in stages:
				labeldict = {0: 'W', 4: 'REM', 1: 'S1', 2: 'S2', 3: 'SWS', }
			else:
				labeldict = {0: 'W', 4: 'REM', 2: 'S2', 3: 'SWS'}
		elif np.max(stages) == 9 and np.min(stages) == 0:
			if 1 in stages:
				labeldict = {0: 'W', 4: 'REM', 1: 'S1', 2: 'S2', 3: 'SWS', 9: 'A'}
			else:
				labeldict = {0: 'W', 4: 'REM', 2: 'S2', 3: 'SWS'}
		else:
			if verbose:
				print('could not detect labels?')
			if 1 in stages:
				labeldict = {0: 'W', 4: 'REM', 1: 'S1', 2: 'S2', 3: 'SWS', 5: 'A'}
			else:
				labeldict = {0: 'W', 4: 'REM', 2: 'S2', 3: 'SWS', 5: 'A'}
		if -1 in stages:
			labeldict['ARTEFACT'] = -1
		if verbose:
			print('Assuming {}'.format(labeldict))

	# check if all stages that are in the hypnogram have a corresponding label in the dict
	for stage in np.unique(stages):
		if stage not in labeldict:
			print('WARNING: {} is in stages, but not in labeldict, stage will be ??'.format(stage))

	# create the label order
	labels = [labeldict[label] for label in labeldict]
	labels = sorted(set(labels), key=labels.index)

	# we iterate through the stages and fetch the label for this stage
	# then we append the position on the plot of this stage via the labels-dict
	x = []
	y = []
	rem_start = []
	rem_end = []
	for i in np.arange(len(stages)):
		s = stages[i]
		label = labeldict.get(s)
		if label is None:
			p = 99
			if '??' not in labels:
				labels.append('??')
		else:
			p = -labels.index(label)

		# make some red line markers for REM, mark beginning and end of REM
		if 'REM' in labels:
			if label == 'REM' and len(rem_start) == len(rem_end):
				rem_start.append(i - 2)
			elif label != 'REM' and len(rem_start) > len(rem_end):
				rem_end.append(i - 1)
		if label == 'REM' and i == len(stages) - 1:
			rem_end.append(i + 1)

		if i != 0:
			y.append(p)
			x.append(i - 1)
		y.append(p)
		x.append(i)

	assert len(rem_start) == len(rem_end), 'Something went wrong in REM length calculation'

	x = np.array(x) * epochlen
	y = np.array(y)
	y[y == 99] = y.min() - 1  # make sure Unknown stage is plotted below all else

	if ax is None:
		plt.figure()
		ax = matplotlib.pyplot.gca()
	formatter = matplotlib.ticker.FuncFormatter(lambda s, x: time.strftime('%H:%M', time.gmtime(s)))

	ax.plot(x, y, **kwargs)
	ax.set_xlim(0, x[-1])
	ax.xaxis.set_major_formatter(formatter)

	ax.set_yticks(np.arange(len(np.unique(labels))) * -1)
	ax.set_yticklabels(labels)
	ax.set_xticks(np.arange(0, x[-1], 3600))
	if xlabel:
		plt.xlabel('Time after recording start')
	if ylabel:
		plt.ylabel('Sleep Stage')
	if title is not None:
		ax.set_title(title)

	try:
		warnings.filterwarnings("ignore", message='This figure includes Axes that are not compatible')
		plt.tight_layout()
	except Exception:
		pass

	# plot REM in RED here
	for start, end in zip(rem_start, rem_end):
		height = -labels.index('REM')
		ax.hlines(height, start * epochlen, end * epochlen, color='r', linewidth=4, zorder=99)



import pandas as pd
import pyedflib as edf
from typing import List


def get_edf(file) -> (List[float], pd.DataFrame):
	"""

	:param file: The edf file to be read
	:return: An array of all the signals and a DataFrame with information about the signals
	"""

	# Read the file and return the signals, data about the signals and data about the file
	signals, signal_headers, header = edf.highlevel.read_edf(file)

	# Turn the signals headers into a DataFrame
	signal_headersdf = pd.DataFrame(data=signal_headers)

	return signals, signal_headersdf


# if __name__ == "__main__":
# 	file = "C:\Haskoli\YearThree\Spring\LOKA\arora\src\ma0844az_1-1+.edf"
# 	get_edf(file)



import pandas as pd
from scipy.interpolate import interp1d
from numpy import sign, zeros

# from scipy.interpolate import CubicSpline
# from numpy import array


def upper_lower_signal_envelopes(epoch: pd.DataFrame):
	"""

	Args:
		epoch: array of signals that has been segmented from a larger array -> A DataFrame

	Returns: Two numpy arrays that represent the upper and lower

	"""
	s = epoch
	q_u = zeros(s.shape)
	q_l = zeros(s.shape)

	# Prepend the first value of (s) to the interpolating values. This forces the model to use the same starting point
	# for both the upper and lower envelope models.

	u_x = [0, ]
	u_y = [s[0], ]

	l_x = [0, ]
	l_y = [s[0], ]

	# Detect peaks and troughs and mark their location in u_x,u_y,l_x,l_y respectively.

	for k in range(1, len(s) - 1):
		if (sign(s[k] - s[k - 1]) == 1) and (sign(s[k] - s[k + 1]) == 1):
			u_x.append(k)
			u_y.append(s[k])

		if (sign(s[k] - s[k - 1]) == -1) and ((sign(s[k] - s[k + 1])) == -1):
			l_x.append(k)
			l_y.append(s[k])

	# Append the last value of (s) to the interpolating values. This forces the model to use the same ending point for
	# both the upper and lower envelope models.

	u_x.append(len(s) - 1)
	u_y.append(s[-1])

	l_x.append(len(s) - 1)
	l_y.append(s[-1])

	# Fit suitable models to the data. Here I am using cubic splines,
	# similarly to the MATLAB example given in the question.

	u_p = interp1d(u_x, u_y, kind='cubic', bounds_error=False, fill_value=0.0)
	l_p = interp1d(l_x, l_y, kind='cubic', bounds_error=False, fill_value=0.0)
	# u_p = CubicSpline(u_x,u_y)
	# l_p = CubicSpline(l_x,l_y)
	# Evaluate each model over the domain of (s)
	for k in range(0, len(s)):
		q_u[k] = u_p(k)
		q_l[k] = l_p(k)
	return q_u, q_l



from random import random
from typing import List
import numpy as np
from scipy import signal


def segment_fs(signals: List[float or int], fs: int) -> List[List[int or float]]:
	"""

	Args:
		signals:
		fs:

	Returns:

	"""
	# Maybe have this in a separate file for everyone to use
	return [signals[x:x + fs] for x in range(0, len(signals), fs)]


def resample(signals: List[int or float], newfs: int, oldfs: int):
	"""
	Matias gave us the base of this code
	Args:
		signals:
		newfs:
		oldfs:

	Returns:

	"""
	return signal.resample(signals, int((newfs/oldfs)*len(signals)))


def lower_frequency(signals: List[float or int], method: str, new_fs: int) -> List[float or int]:
	"""
	Lowers the frequency of the given signals by some given method
	Args:
		signals: List of time series signals
		method: The method how the frequency should be lowered
		new_fs: The new frequency of the signals

	Returns: List of new values corresponding to the method given

	"""
	new_signal = []
	segmented_signal = segment_fs(signals, new_fs)
	if method.lower() == 'min':
		for index in range(0, len(segmented_signal)):
			the_lowest = min(segmented_signal[index])
			new_signal.append(the_lowest)

	elif method.lower() == 'max':
		for index in range(0, len(segmented_signal)):
			the_lowest = max(segmented_signal[index])
			new_signal.append(the_lowest)

	elif method.lower() == 'mean':
		for index in range(0, len(segmented_signal)):
			the_lowest = np.mean(segmented_signal[index])
			new_signal.append(the_lowest)

	elif method.lower() == 'median':
		for index in range(0, len(segmented_signal)):
			the_lowest = np.median(segmented_signal[index])
			new_signal.append(the_lowest)

	return new_signal



from scipy import signal
import numpy as np
from typing import List

"""
TODO: Make the code more readable and document/comment it
"""


def standardize(epoch: List[int or float]):
	"""
	Matias gave us this code
	Z-score normalization for the signals.
	Args:
		epoch:
	Returns:
	"""
	return (epoch - np.mean(epoch)) / np.std(epoch)


def low_pass_filter(
		data: List[int or float],
		cutoff: int or float,
		frequency: int or float,
		order=5) -> np.ndarray:
	"""
	:param data: Array of integers or float
	:param cutoff: The frequency that is to be filtered through --> integer or float
	:param frequency: The frequency of the data --> integer or float
	:param order: The order of the filter --> integer
	:return: An array of filtered data using the Buttersworth method --> Numpy ndarray
	"""
	b, a = _butter_pass(cutoff, frequency, btype='lowpass', order=order)
	y = signal.lfilter(b, a, data)
	return y


def cheby2_highpass_filtfilt(
		signals,
		fs,
		cutoff,
		order=5,
		rs=40.0):
	"""
	Chebyshev type1 highpass filtering.
	Matias gave us this code
	Args:
		signals: the signals
		fs: sampling freq in Hz
		cutoff: cutoff freq in Hz
		order:
		rs:
	Returns:
		the filtered signals
	"""
	nyq = 0.5 * fs
	norm_cutoff = cutoff / nyq
	sos = signal.cheby2(order, rs, norm_cutoff, btype='highpass', output='sos')
	return signal.sosfiltfilt(sos, signals)


def high_pass_filter(
		data: List[int or float] or np.ndarray,
		cutoff: int or float,
		frequency: int or float,
		order=5) -> np.ndarray:
	"""
	Args:
		data: Array of integers or float
		cutoff: The frequency that is to be filtered through --> integer or float
		frequency: The frequency of the data --> integer or float
		order: The order of the filter --> integer

	Returns: An array of filtered data using the Buttersworth method --> Numpy ndarray
	"""
	b, a = _butter_pass(cutoff, frequency, btype='highpass', order=order)
	y = signal.filtfilt(b, a, data)
	return y


def _butter_pass(
		cutoff: int or float,
		frequency: int or float,
		btype: str,
		order: int = 5) -> (np.ndarray, np.ndarray):
	"""
	Args:
		cutoff: The frequency that is to be filtered through --> integer or float
		frequency: The frequency of the data --> integer or float
		btype: The order of the filter --> integer
		order: The type of filtering --> string
	Returns:The denominator and the numerator polynomials of the IIR filter --> Numpy ndarray, numpy ndarray
	"""
	nyq = 0.5 * frequency
	normal_cutoff = cutoff / nyq
	b, a = signal.butter(order, normal_cutoff, btype=btype, analog=False)
	return b, a



import signal
from typing import Tuple
import pandas as pd
import numpy as np
from arora.filter.EEG_freqbands import eeg_freq_bands


def welch_psd(epoch: pd.DataFrame, fs: int or float, filter_order: int) \
		-> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
	"""

	:param epoch: a segment of an edf signals in the form of a 1d array
	:param fs: the frequency of the epoch in the form of an integer
	:param filter_order: ???
	:return: ???
	"""
	# Acknowledgement: Katrín Hera, M.Sc
	delta_band, theta_band, alpha_band, beta_band = eeg_freq_bands(epoch, fs, filter_order)

	window_size = int((fs / 2) - 1)
	delta_nonbiased = delta_band - signal.savgol_filter(delta_band, window_size, filter_order)
	theta_nonbiased = theta_band - signal.savgol_filter(theta_band, window_size, filter_order)
	alpha_nonbiased = alpha_band - signal.savgol_filter(alpha_band, window_size, filter_order)
	beta_nonbiased = beta_band - signal.savgol_filter(beta_band, window_size, filter_order)
	F_AF3, delta_PSD = signal.welch(delta_nonbiased, fs / 2, nperseg=len(delta_nonbiased))
	F_AF3, theta_PSD = signal.welch(theta_nonbiased, fs / 2, nperseg=len(theta_nonbiased))
	F_AF3, alpha_PSD = signal.welch(alpha_nonbiased, fs / 2, nperseg=len(alpha_nonbiased))
	F_AF3, beta_PSD = signal.welch(beta_nonbiased, fs / 2, nperseg=len(beta_nonbiased))

	return delta_PSD, theta_PSD, alpha_PSD, beta_PSD



import numpy as np
import pandas as pd

from arora import math
from arora import upper_lower_signal_envelopes


def find_features(epochized_signal: pd.DataFrame, sampling_freq: int or float,
                  filter_order: int) -> pd.DataFrame:
    """
    Args:
        epochized_signal: The segmented EEG data -> A DataFrame
        sampling_freq: The sampling frequency of the epochized signals -> an integer or a float
        filter_order: What the filter order should be -> integer

    Returns: A DataFrame that contains the main features of the signals

    """

    l = len(epochized_signal)
    mini = []
    maxi = []
    med = []
    upper_mean = []
    lower_mean = []
    upper_std = []
    lower_std = []
    upper_var = []
    lower_var = []
    PSD = []

    for i in range(0, l):
        mini.append(np.min(epochized_signal[i]))
        maxi.append(np.max(epochized_signal[i]))
        med.append(np.median(epochized_signal[i]))

        upper, lower = upper_lower_signal_envelopes(epochized_signal[i])
        upper_mean.append(np.mean(upper))
        lower_mean.append(np.mean(lower))
        upper_std.append(np.std(upper))
        lower_std.append(np.std(lower))
        upper_var.append(np.var(upper))
        lower_var.append(np.var(lower))

        PSDlist = math.welch_psd(epochized_signal[i], sampling_freq, filter_order)
        PSD.append(PSDlist)

    dataframe = epochized_signal

    dataframe['Epoch min'] = mini
    dataframe['Epoch max'] = maxi
    dataframe['Epoch median'] = med
    dataframe['Upper env mean'] = upper_mean
    dataframe['Lower env mean'] = lower_mean
    dataframe['Upper env std'] = upper_std
    dataframe['Lower env std'] = lower_std
    dataframe['Upper env var'] = upper_var
    dataframe['Lower env var'] = lower_var
    dataframe['PSD'] = PSD

    return dataframe



from typing import List

from numpy import ndarray

import arora.filter.bandpass


def eeg_freq_bands(eeg_signal: List[int or float], sampling_freq: float or int,
                   filter_order: float or int) -> (ndarray, ndarray, ndarray, ndarray):
	"""

	Args:
		eeg_signal: One type of signals from the edf file
		sampling_freq: The sampling frequency of the eeg signals
		filter_order:

	Returns: Four arrays that contain each bandpass of the most common frequencies

	"""
	alpha_lower = 8
	alpha_upper = 14
	beta_lower = 14
	beta_upper = 50
	delta_lower = 0.5
	delta_upper = 4
	theta_lower = 4
	theta_upper = 8
	delta_band = arora.bandpass(eeg_signal, delta_lower, delta_upper, sampling_freq, filter_order)
	theta_band = arora.bandpass(eeg_signal, theta_lower, theta_upper, sampling_freq, filter_order)
	alpha_band = arora.bandpass(eeg_signal, alpha_lower, alpha_upper, sampling_freq, filter_order)
	beta_band = arora.bandpass(eeg_signal, beta_lower, beta_upper, sampling_freq, filter_order)
	return delta_band, theta_band, alpha_band, beta_band



from typing import List

import numpy as np
import scipy
from scipy import signal

"""
This is a work in a WIP and will be changed, at the moment is butter returned too many values so 
this will be needed to change
"""


def bandpass(raw_signal: List[int or float], lower_cutoff: int or float,
             upper_cutoff: int or float, sampling_freq: int or float, filter_order: int) -> np.ndarray:
    """

	Args:
		raw_signal: The EEG signals that is to be processed -> an array of float/integers value
		lower_cutoff: The lower frequency value -> a float or an integer
		upper_cutoff: The upper frequency value -> a float or an integer
		sampling_freq: The frequency of the raw_signal -> a float or an integer
		filter_order: The maximum number of delay elements used in the filter circuit -> an integer value

	Returns: A filtered array

	"""

	w1 = lower_cutoff / (sampling_freq / 2) # Normalize the frequencies of the two
	w2 = upper_cutoff / (sampling_freq / 2)

    # w1 = lower_cutoff / sampling_freq
    # w2 = upper_cutoff / sampling_freq
    b, a = signal.butter(filter_order, [w1, w2], btype='bandpass', output='ba') #make filter
    output_band = scipy.signal.lfilter(b, a, raw_signal)
    return output_band



import setuptools
import os

with open("README.md", "r") as fh:
    long_description = fh.read()
    
package_name = os.getenv(key="PACKAGE_NAME", default="arora")

setuptools.setup(
    name=package_name,
    version=os.getenv(key="PACKAGE_VERSION", default="0.1.0"),
    long_description=long_description,
    long_description_content="text/markdown",
    author="Sleep Revolution",
    author_email="sleeprevolution@ru.is",
    description="A tool for sleep researcher to preprocess, work, analyze and visualize data",
    packages=setuptools.find_packages(include=[package_name, f"{package_name}.*"]),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8"
        ],
    install_requires=["numpy", "pandas", "pyEDFlib", "scipy"],
    python_requires=">=3.8"
)



from typing import List

import numpy as np
import statistics


def iqr_standardize(signals: List[int or float], lowerquantile: int, upperquantile: int):
	"""
	IQR standardization for the signals.
	Args:
		signals:
		lowerquantile:
		upperquantile:

	Returns:

	"""
	upperq, lowerq = np.percentile(signals, [upperquantile, lowerquantile])
	iqr = upperq - lowerq
	return (signals - np.median(signals)) / iqr


def mean_val(signal_list: List[int or float]) -> float or int:
	"""
	Function that find the mean of a give array

	:param signal_list: a 1d array of edf signals
	:return: The mean of the array
	"""
	length = len(signal_list)
	if length == 0:
		return 0.0

	return float(sum(signal_list) / length)


def std_val(signal_list: List[int or float]) -> float:
	"""
	Function that finds the given standard deviation of a given array

	:param signal_list:  a 1d array of edf signals
	:return: The standard deviation of the array
	"""
	if len(signal_list) < 2:
		raise Exception("Your array must contain at least two values")
	return statistics.stdev(signal_list)



"""
THIS IS A WIP file and is subject to change, will probably be turned into a directory later on
when more features have been implemented
"""
import datetime
import pandas as pd
from pandas import DataFrame
from typing import Union, List
import numpy as np


# TODO: comment
# TODO: pretty up the code through typing conventions
# TODO: UnitTests


def segmentation(signals: List[int or float], freq: float, onset: int,
                 duration: int = 30, time_unit: str = "sec") -> (List[int or float], pd.DataFrame):
    """
    TODO: Allow the user to decide if he only wants to get the signals or also the Dataframe
    :param onset: start of recording in seconds (int)
    :param signals: list of EEG signals (float)
    :param freq: sample frequency of the EEG signals (int)
    :param duration: end of recording in seconds (int)
    :param time_unit: the unit of time for the duration (str)
    :return: a list of EEG signals and pandas DataFrame including the beginning and end index of the sample
    """
    if time_unit == "min":
        duration *= 60
    elif time_unit == "hours":
        duration *= 3600

    signal_list = []
    index_dataframe = pd.DataFrame()
    temp_dataframe = pd.DataFrame()

    # find the beginning index by transferring time into index using the sampling frequency
    beg_index = int(onset * freq)

    # find the end index by figuring out the
    end_index = int((onset + duration) * freq)

    # end index might be out of scope of the list in which case we'll return as much as we can
    if end_index > len(signals):
        # TODO: warning/error handling over here
        print("Warning: Duration is beyond the scope of your list, data might not be reliable")

        signal_list = signals[beg_index:]
        end_index = len(signals)
        index_dataframe['beg_index'] = [beg_index]
        index_dataframe['end_index'] = [end_index]

        return signal_list, index_dataframe

    while end_index <= len(signals):
        # for _ in range(beg_index, len(signals), end_index):
        signal_list.append(signals[beg_index:end_index + 1])
        temp_dataframe['beg_index'] = [beg_index]
        temp_dataframe['end_index'] = [end_index]
        index_dataframe = pd.concat([temp_dataframe, index_dataframe], axis=0, sort=True, ignore_index=True)

        onset += duration
        beg_index = end_index + 1
        end_index = int((onset + duration) * freq)

        # index_dataframe = pd.concat([temp_dataframe, index_dataframe], sort=False)

    return signal_list, index_dataframe


def epochize(data: List[int or float], channel_names: List[str], epoch_len: int,
             sampling_freq: int or float, start_timestamp: datetime) -> pd.DataFrame:
    """
    Args:
        data: EDF file turned into an array -> pd.DataFrame
        channel_names: The names of the channels to be used -> list[str]
        epoch_len: How long the epoch should be - will be changed to duration later on -> integer
        sampling_freq: The sampling frequency of the data -> float or integer
        start_timestamp: ??? - Will be removed later on

    Returns: A dataframe with the epochs - will be changed later on to a tuple with the epoch,
            start and duration
    """
    # WIP, will need to change the return, and some things in the function so that it will fit
    # previous work
    # Acknowledgement: Katrín Hera, M.Sc. student

    # Find the length of the signals
    l = len(data)
    # Find the number of data points
    number = int(epoch_len * sampling_freq)
    # initialize the epoch list
    epochs = []
    # add the epochs to the list
    for x in range(0, l, int(epoch_len * sampling_freq)):
        epochs.append(data[x:x + number])
    # initialize a list of timestamps at the beginning of each epoch
    timestamps_epoch_start = []
    for i in range(int(len(data) / (epoch_len * sampling_freq))):
        # make start timestamps for all the subsequent epochs and add to list
        next_timestamp = start_timestamp + datetime.timedelta(seconds=epoch_len * i)
        timestamps_epoch_start.append(next_timestamp)
    # make dataframe with timestamps and epoch
    df: DataFrame = pd.DataFrame()
    df['Epoch start'] = timestamps_epoch_start
    df[channel_names] = epochs  # temporarilly commented out for debugging

    # df['eh_test'] = pd.to_datetime(df['start times'])
    # df = df.set_index('Epoch start')
    return df


"""
TODO: Implement the functions and add them to the adaptive segmentations function
TODO: Document and comment the functions
TODO: Move find_transients function to some other file
"""


def _signal_afc(N: Union[int, float], i: Union[int, float],
                signals: Union[List[Union[int, float]], np.ndarray]):
    R_sum = 0

    for n in range(1, N - i):
        R_sum += signals[n] * signals[n + i]

    R_sum = R_sum * (1 / N)

    return R_sum


def _lp_filter_creation():
    return None


def _est_signal_value(signal, p, a, n):
    the_sum = 0
    for index in range(1, p):
        # TODO: check if the n is smaller than p and make some error handling for that, this should be able to go back in time
        the_sum += a[index] * signal[n - index]
    return -1 * the_sum


def error_of_signal():
    return


def _calculate_pe_val():
    return None


def _calculate_pe_acf():
    return None


def _calculate_sem():
    return None


def find_transients(hello):
    # This is part of the feature extraction process
    return None


def adaptive_segmentation(signals: Union[List[Union[int, float]], np.ndarray]):
    """
    :param signals:
    :return:
    """
    # Store segment length - Store both the onset and the duration
    # Store signals predictor
    # Store corrective predictor
    N = None or 10  # maybe 50?
    i = None or 666 or 56

    # Step 1:
    # 	a. Compute new signals ACF
    _signal_afc(N, i, signals)

    # 	b. adapt LP filter
    # Step 2:

    return None
